check_recipe <- function(recipe) {

  # recipe: yaml recipe already read it
  ## TODO: Adapt to decadal case

  info(recipe$Run$logger, paste("Checking recipe:", recipe$recipe_path))

  # ---------------------------------------------------------------------
  # ANALYSIS CHECKS
  # ---------------------------------------------------------------------

  TIME_SETTINGS_SEASONAL <- c("sdate", "ftime_min", "ftime_max", "hcst_start",
			      "hcst_end")
  TIME_SETTINGS_DECADAL <- c("ftime_min", "ftime_max", "hcst_start", "hcst_end") 
  PARAMS <- c("Horizon", "Time", "Variables", "Region", "Regrid", "Workflow",
	      "Datasets")
  HORIZONS <- c("subseasonal", "seasonal", "decadal")
  # Define error status variable
  error_status <- F

  # Check basic elements in recipe:Analysis:
  if (!("Analysis" %in% names(recipe))) {
    error(recipe$Run$logger,
	  "The recipe must contain an element called 'Analysis'.")
    error_status <- T
  }

  if (!all(PARAMS %in% names(recipe$Analysis))) {
    error(recipe$Run$logger, 
          paste0("The element 'Analysis' in the recipe must contain all of ",
		 "the following: ", paste(PARAMS, collapse = ", "), "."))
    error_status <- T
  }

  if (!any(HORIZONS %in% tolower(recipe$Analysis$Horizon))) {
    error(recipe$Run$logger,
          paste0("The element 'Horizon' in the recipe must be one of the ",
		"following: ", paste(HORIZONS, collapse = ", "), "."))
    error_status <- T
  }
  # Check time settings
  if (tolower(recipe$Analysis$Horizon) == "seasonal") {
    if (!all(TIME_SETTINGS_SEASONAL %in% names(recipe$Analysis$Time))) {
      error(recipe$Run$logger,
	    paste0("The element 'Time' in the recipe must contain all of the ",
		   "following: ", paste(TIME_SETTINGS_SEASONAL,
					collapse = ", "), "."))
      error_status <- T
    }
  } else if (tolower(recipe$Analysis$Horizon) == "decadal") {
    if (!all(TIME_SETTINGS_DECADAL %in% names(recipe$Analysis$Time))) {
      error(recipe$Run$logger,
	    paste0("The element 'Time' in the recipe must contain all of the ",
		   "following: ", paste(TIME_SETTINGS_DECADAL,
					collapse = ", "), "."))
      error_status <- T
    }
  }
  # Check ftime_min and ftime_max
  if ((!(recipe$Analysis$Time$ftime_min > 0)) ||
      (!is.integer(recipe$Analysis$Time$ftime_min))) {
    error(recipe$Run$logger,
	  "The element 'ftime_min' must be an integer larger than 0.")
    error_status <- T
  }
  if ((!(recipe$Analysis$Time$ftime_max > 0)) ||
      (!is.integer(recipe$Analysis$Time$ftime_max))) {
    error(recipe$Run$logger,
	  "The element 'ftime_max' must be an integer larger than 0.")
    error_status <- T
  }
  if (recipe$Analysis$Time$ftime_max < recipe$Analysis$Time$ftime_min) {
    error(recipe$Run$logger,
	  "'ftime_max' cannot be smaller than 'ftime_min'.")
    error_status <- T
  }
  # Check consistency of hindcast years
  if (!(as.numeric(recipe$Analysis$Time$hcst_start) %% 1 == 0) ||
      (!(recipe$Analysis$Time$hcst_start > 0))) {
    error(recipe$Run$logger,
	  "The element 'hcst_start' must be a valid year.")
    error_status <- T
  }
  if (!(as.numeric(recipe$Analysis$Time$hcst_end) %% 1 == 0) ||
      (!(recipe$Analysis$Time$hcst_end > 0))) {
    error(recipe$Run$logger,
	  "The element 'hcst_end' must be a valid year.")
    error_status <- T
  }
  if (recipe$Analysis$Time$hcst_end < recipe$Analysis$Time$hcst_start) {
    error(recipe$Run$logger,
	  "'hcst_end' cannot be smaller than 'hcst_start'.")
    error_status <- T
  }
  ## TODO: Is this needed?
  if (is.null(recipe$Analysis$Time$fcst_year) || 
      tolower(recipe$Analysis$Time$fcst_year) == 'none') {
    stream <- "hindcast"
    # recipe$Analysis$Time$fcst_year <- 'YYYY'
  } else {
    stream <- "fcst"
  }

  ## TODO: To be implemented in the future
  # if (length(recipe$Analysis$Time$sdate$fcst_day) > 1 &&
  #     tolower(recipe$Analysis$Horizon) != "subseasonal") {
  #   warn(recipe$Run$logger,
  #        paste("Only subseasonal verification allows multiple forecast days."),
  #              "Element fcst_day in recipe set as 1.")
  #   recipe$Analysis$Time$sdate$fcst_day <- '01'
  # }
  ## TODO: Delete, this parameter was deprecated
  # if (is.null(recipe$Analysis$Time$sdate$fcst_sday)) {
  #   error(recipe$Run$logger,
  #         paste("The element 'fcst_sday' in the recipe should be defined."))
  # }

  if (is.null(recipe$Analysis$Time$fcst_year)) {
    warn(recipe$Run$logger,
         paste("The element 'fcst_year' is not defined in the recipe.",
	       "No forecast year will be used."))
  }
  ## TODO: Adapt and move this inside 'if'? 
  # fcst.sdate <- NULL
  # for (syear in recipe$Analysis$Time$fcst_year) {
  #   for (sday in recipe$Analysis$Time$sdate) {
  #     fcst.sdate <- c(fcst.sdate, 
  #                     paste0(syear,
  #                            sprintf("%04d", as.numeric(sday))))
  #   }
  # }
  # fcst.sdate <- list(stream = stream, fcst.sdate = fcst.sdate)

  # Regrid checks:
  if (length(recipe$Analysis$Regrid) != 2) {
    error(recipe$Run$logger,
          "The 'Regrid' element must specify the 'method' and 'type'.")
    error_status <- T
  }
  # TODO: Add Workflow checks?
  # ...
  # calculate number of workflows to create for each variable and
  if (length(recipe$Analysis$Horizon) > 1) {
    error(recipe$Run$logger,
	  "Only one single Horizon can be specified in the recipe")
    error_status <- T
  }

  ## TODO: Refine this
  # nvar <- length(recipe$Analysis$Variables)
  # if (nvar > 2) {
  #   error(recipe$Run$logger,
  #         "Only two type of Variables can be listed: ECVs and Indicators.")
  #   stop("EXECUTION FAILED")
  # }

  # remove NULL or None Indicators or ECVs from the recipe:
  if (!is.null(recipe$Analysis$Variables$Indicators) &&
      !is.list(recipe$Analysis$Variables$Indicators)) {
    recipe$Analysis$Variables <- recipe$Analysis$Variables[
           -which(names(recipe$Analysis$Variables) == 'Indicators')]
  }
  if (!is.null(recipe$Analysis$Variables$ECVs) &&
      !is.list(recipe$Analysis$Variables$ECVs)) {
    recipe$Analysis$Variables <- recipe$Analysis$Variables[
           -which(names(recipe$Analysis$Variables) == 'ECVs')]
  }

  # Region checks:
  LIMITS <- c('latmin', 'latmax', 'lonmin', 'lonmax')
  for (region in recipe$Analysis$Region) {
    if (!all(LIMITS %in% names(region))) {
      error(recipe$Run$logger,
            paste0("There must be 4 elements in 'Region': ",
		   paste(LIMITS, collapse = ", "), "."))
      error_status <- T
    }
  }
  if (length(recipe$Analysis$Region) > 1) {
    for (region in recipe$Analysis$Region) {
      if (is.null(region$name)) {
	error(recipe$Run$logger,
	      paste("If more than one region has been defined, every region",
		    "must have a unique name."))
      }
    }
  }
    ## TODO: Implement multiple regions
    # nregions <- length(recipe$Analysis$Region)
    # for (i in 1:length(recipe$Analysis$Region)) {
    #   if (!all(limits %in% names(recipe$Analysis$Region[[i]]))) {
    #     limits <- paste(limits, collapse = " ")
    #     error(recipe$Run$logger,
    #           paste0("Each region defined in element 'Region' ",
    #                  "should have 4 elements: ",
    #                  paste(limits, collapse = ", "), "."))
    #     error_status <- T
    #   }
    #   if (length(recipe$Analysis$Region) > 1) {
    #     if (!("name" %in% names(recipe$Analysis$Region[[i]]))) {
    #       error(recipe$Run$logger,
    #             paste("If multiple regions are requested, each region must",
    #                   "have a 'name'".)
    #   # are numeric? class list mode list 
    # }
  # ---------------------------------------------------------------------
  # WORKFLOW CHECKS
  # ---------------------------------------------------------------------

  # Only one Calibration method allowed:
  if ((is.logical(recipe$Analysis$Workflow$Calibration$method) &&
       recipe$Analysis$Workflow$Calibration$method == FALSE) ||
      tolower(recipe$Analysis$Workflow$Calibration$method) == 'none' ||
      is.null(recipe$Analysis$Workflow$Calibration$method)) {
    warn(recipe$Run$logger,
         "No Calibration method was specified, raw data verification.")
    recipe$Analysis$Workflow$Calibration$method <- 'raw'
  } else {
    if (is.null(recipe$Analysis$Workflow$Calibration$method)) {
      error(recipe$Run$logger,
            "The 'Calibration' element 'method' must be specified.")
      error_status <- T
    }
  }
  # Anomalies
  if ("Anomalies" %in% names(recipe$Analysis$Workflow)) {
    if (is.null(recipe$Analysis$Workflow$Anomalies$compute)) {
      error(recipe$Run$logger,
	    "Parameter 'compute' must be defined under 'Anomalies'.")
      error_status <- T
    } else if (!(is.logical(recipe$Analysis$Workflow$Anomalies$compute))) {
      error(recipe$Run$logger,
	    paste("Parameter 'Anomalies:compute' must be a logical value",
		  "(True/False or yes/no)."))
      error_status <- T
    } else if ((recipe$Analysis$Workflow$Anomalies$compute) &&
	       (!is.logical(recipe$Analysis$Workflow$Anomalies$cross_validation))) {
      error(recipe$Run$logger,
	    paste("If anomaly computation is requested, parameter",
		  "'cross_validation' must be defined under 'Anomalies',
		  and it must be a logical value (True/False or yes/no)."))
      error_status <- T
    }
  }
  # Skill
  if (("Skill" %in% names(recipe$Analysis$Workflow)) &&
      (is.null(recipe$Analysis$Workflow$Skill$metric))) {
    error(recipe$Run$logger,
	  "Parameter 'metric' must be defined under 'Skill'.")
    error_status <- T
  }
  # Probabilities
  if ("Probabilities" %in% names(recipe$Analysis$Workflow)) {
    if (is.null(recipe$Analysis$Workflow$Probabilities$percentiles)) {
      error(recipe$Run$logger,
	    "Parameter 'percentiles' must be defined under 'Probabilities'.")
      error_status <- T
    } else if (!is.list(recipe$Analysis$Workflow$Probabilities$percentiles)) {
      error(recipe$Run$logger,
	    paste("Parameter 'Probabilities:percentiles' expects a list.",
		  "See documentation in the wiki for examples."))
      error_status <- T
    }
  }

  # ---------------------------------------------------------------------
  # RUN CHECKS
  # ---------------------------------------------------------------------

  RUN_FIELDS = c("Loglevel", "Terminal", "output_dir", "code_dir")
  LOG_LEVELS = c("INFO", "DEBUG", "WARN", "ERROR", "FATAL")

  if (!("Run" %in% names(recipe))) {
    stop("The recipe must contain an element named 'Run'.")
  }
  if (!all(RUN_FIELDS %in% names(recipe$Run))) {
    error(recipe$Run$logger, paste("Recipe element 'Run' must contain",
				   "all of the following fields:",
                                   paste(RUN_FIELDS, collapse=", "), "."))
    error_status <- T
  }
  if (!is.character(recipe$Run$output_dir)) {
    error(recipe$Run$logger,
          paste("The Run element 'output_dir' in", recipe$name, "file",
                "should be a character string indicating the path where",
                "the outputs should be saved."))
    error_status <- T
  }
  if (!is.character(recipe$Run$code_dir)) {
    error(recipe$Run$logger,
          paste("The Run element 'code_dir' in", recipe$name, "file ",
                "should be a character string indicating the path",
                "where the code is."))
    error_status <- T
  }
  if (!is.logical(recipe$Run$Terminal)) {
    error(recipe$Run$logger,
          paste("The Run element 'Terminal' in", recipe$name, "file ",
                "should be a boolean value indicating whether or not to",
                "print the logs in the terminal."))
    error_status <- T
  }
  ## TODO: Review this case, since default value is allowed
  if (!is.character(recipe$Run$Loglevel) || 
      !any(recipe$Run$Loglevel %in% LOG_LEVELS)) {
    error(logger,
          paste("The Run element 'Loglevel' in", recipe$name, "file",
                "should be a character string specifying one of the levels available:",
                paste0(LOG_LEVELS, collapse='/')))
    error_status <- T
  }

  # ---------------------------------------------------------------------
  # WORKFLOW CHECKS
  # ---------------------------------------------------------------------
  # Check workflow: need to define restrictions? 
  # e.g. only one calibration method
  ## TODO: Implement number of dependent verifications
  #nverifications <- check_number_of_dependent_verifications(recipe)
  # info(recipe$Run$logger, paste("Start Dates:",
  #  				  paste(fcst.sdate, collapse = " ")))

  # Return error if any check has failed
  if (error_status) {
    error(recipe$Run$logger, "RECIPE CHECK FAILED.")
    stop("The recipe contains some errors. Find the full list in the",
	 "startup.log file.")
  } else {
    info(recipe$Run$logger, "##### RECIPE CHECK SUCCESSFULL #####")
    # return(append(nverifications, fcst.sdate))
  }
}

check_number_of_dependent_verifications <- function(recipe) {
  # Number of verifications depends on the variables and indicators requested
  # and the order of the workflow:
  # workflow: correction + indicator --> only 1 variable is calibrated
  # workflow: indicator + correction --> the indicator and the ecv are calibrated
  independent_verifications <- NULL
  dependent_verifications <- NULL
  dep <- 1
  # check workflow order:
  if (all(c('Calibration', 'Indicators') %in% names(recipe$Analysis$Workflow))) {
    cal_pos <- which(names(recipe$Analysis$Workflow) == 'Calibration')
    ind_pos <- which(names(recipe$Analysis$Workflow) == 'Indicators')
    if (cal_pos < ind_pos) {
      workflow_independent <- FALSE
    } else {
      workflow_independent <- TRUE
    }
  }
  if (workflow_independent) {
    independent_verifications <- append(recipe$Analysis$Variables$ECVs,
                                        recipe$Analysis$Variables$Indicators)
  } else { 
    if (is.null(recipe$Analysis$Variables$Indicators) || 
       (length(recipe$Analysis$Variables$Indicators) == 1 && 
        is.null(recipe$Analysis$Variables$ECVs))) {
      independent_verifications <- append(recipe$Analysis$Variables$ECVs,
                                          recipe$Analysis$Variables$Indicators)
    } else { 
      ecvs <- recipe$Analysi$Variables$ECVs
      inds <- recipe$Analysi$Variables$Indicators
      ind_table <- read_yaml(paste0(recipe$Run$code_dir,
                             "conf/indicators_table.yml"))
      # first, loop on ecvs if any and compare to indicators
      done <- NULL # to gather the indicators reviewed
      if (!is.null(ecvs)) {
        for (i in 1:length(ecvs)) {
          dependent <- list(ecvs[[i]])
          for (j in 1:length(inds)) {
            if (ind_table[inds[[j]]$name][[1]]$ECVs == ecvs[[i]]$name) {
              if (ind_table[inds[[j]]$name][[1]]$freq == ecvs[[i]]$freq) {
              # they are dependent
                dependent <- append(dependent, inds[[j]])
                done <- append(done, inds[[j]])
              }
            }
          }
          if (length(dependent) == 1) {
            dependent <- NULL
            independent_verifications <- append(independent_verifications,
                                              list(ecvs[[i]]))
          } else {
            dependent_verifications <- append(dependent_verifications,
                                              list(dependent))
          }  
        }
        # There are indicators not reviewed yet?
        if (length(done) < length(inds)) {
          if (length(inds) == 1) {
            independent_verifications <- append(independent_verifications,
                                                inds)
          } else {
            done <- NULL
            for (i in 1:(length(inds) - 1)) {
              dependent <- list(inds[[i]]$name)
              if (is.na(match(unlist(dependent), unlist(done)))) {
                for (j in (i+1):length(inds)) {
                  if (ind_table[inds[[i]]$name][[1]]$ECVs == 
                      ind_table[inds[[j]]$name][[1]]$ECVs) {
                    if (ind_table[inds[[i]]$name][[1]]$freq ==
                        ind_table[inds[[j]]$name][[1]]$freq) {
                      dependent <- append(dependent, inds[[j]]$name)
                      done <- dependent
                    }
                  }
                }
              } 
              if (length(dependent) == 1) {
                independent_verifications <- dependent
                dependent <- NULL     
              } else {
                dependent_verifications <- dependent
              }
            } 
          }
        }
      } else { # there are only Indicators:
        done <- NULL
        for (i in 1:(length(inds) - 1)) {
          dependent <- list(inds[[i]]$name)
          if (is.na(match(unlist(dependent), unlist(done)))) {
            for (j in (i+1):length(inds)) {
              if (ind_table[inds[[i]]$name][[1]]$ECVs ==
                  ind_table[inds[[j]]$name][[1]]$ECVs) {
                if (ind_table[inds[[i]]$name][[1]]$freq ==
                    ind_table[inds[[j]]$name][[1]]$freq) {
                  dependent <- append(dependent, inds[[j]]$name)
                  done <- dependent
                }
              }
            }
          }
          if (length(dependent) == 1) {
            independent_verifications <- dependent
            dependent <- NULL
          } else {
            dependent_verifications <- dependent
          }
        }
      } 
    }
  }
  if (!is.null(independent_verifications)) {   
    info(logger, paste("The variables for independent verification are ", 
         paste(independent_verifications, collapse = " ")))
  }
  if (!is.null(dependent_verifications)) {
    info(logger, paste("The variables for dependent verification are: ",
         paste(dependent_verifications, collapse = " ")))
  }
  # remove unnecessary names in objects to be removed
  return(list(independent = independent_verifications, 
              dependent = dependent_verifications))
}
#workflow <- list(Calibration = list(method = 'SBC'),
#                 Skill = list(metric = 'RPSS'))
#ApplyWorkflow <- function(workflow) {

#res <- do.call('CST_BiasCorrection',
#               args = list(exp = lonlat_data$exp,
#                           obs = lonlat_data$obs))
   



